@@ -143,6 +143,7 @@ enum {
 };
 
 enum {
+  FAILED,
   HANDLE_CALL,
   NOTIFICATION,
   N_SIGNALS
@@ -229,14 +230,17 @@ jsonrpc_client_panic (JsonrpcClient *self,
   g_assert (JSONRPC_IS_CLIENT (self));
   g_assert (error != NULL);
 
-  priv->failed = TRUE;
+  g_object_ref (self);
 
-  jsonrpc_client_close (self, NULL, NULL);
+  priv->failed = TRUE;
 
   /* Steal the tasks so that we don't have to worry about reentry. */
   invocations = g_steal_pointer (&priv->invocations);
   priv->invocations = g_hash_table_new_full (NULL, NULL, NULL, g_object_unref);
 
+  /* Now close the connection */
+  jsonrpc_client_close (self, NULL, NULL);
+
   /*
    * Clear our input and output streams so that new calls
    * fail immediately due to not being connected.
@@ -251,6 +255,10 @@ jsonrpc_client_panic (JsonrpcClient *self,
   g_hash_table_iter_init (&iter, invocations);
   while (g_hash_table_iter_next (&iter, NULL, (gpointer *)&task))
     g_task_return_error (task, g_error_copy (error));
+
+  g_signal_emit (self, signals [FAILED], 0);
+
+  g_object_unref (self);
 }
 
 /*
@@ -306,7 +314,7 @@ jsonrpc_client_constructed (GObject *object)
 }
 
 static void
-jsonrpc_client_finalize (GObject *object)
+jsonrpc_client_dispose (GObject *object)
 {
   JsonrpcClient *self = (JsonrpcClient *)object;
   JsonrpcClientPrivate *priv = jsonrpc_client_get_instance_private (self);
@@ -318,7 +326,7 @@ jsonrpc_client_finalize (GObject *object)
   g_clear_object (&priv->io_stream);
   g_clear_object (&priv->read_loop_cancellable);
 
-  G_OBJECT_CLASS (jsonrpc_client_parent_class)->finalize (object);
+  G_OBJECT_CLASS (jsonrpc_client_parent_class)->dispose (object);
 }
 
 static void
@@ -370,7 +378,7 @@ jsonrpc_client_class_init (JsonrpcClientClass *klass)
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
   object_class->constructed = jsonrpc_client_constructed;
-  object_class->finalize = jsonrpc_client_finalize;
+  object_class->dispose = jsonrpc_client_dispose;
   object_class->get_property = jsonrpc_client_get_property;
   object_class->set_property = jsonrpc_client_set_property;
 
@@ -412,6 +420,26 @@ jsonrpc_client_class_init (JsonrpcClientClass *klass)
 
   g_object_class_install_properties (object_class, N_PROPS, properties);
 
+  /**
+   * JsonrpcClient::failed:
+   *
+   * The "failed" signal is called when the client has failed communication
+   * or the connection has been knowingly closed.
+   *
+   * Since: 3.28
+   */
+  signals [FAILED] =
+    g_signal_new ("failed",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  G_STRUCT_OFFSET (JsonrpcClientClass, failed),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
+  g_signal_set_va_marshaller (signals [FAILED],
+                              G_TYPE_FROM_CLASS (klass),
+                              g_cclosure_marshal_VOID__VOIDv);
+
   /**
    * JsonrpcClient::handle-call:
    * @self: A #JsonrpcClient
@@ -428,12 +456,16 @@ jsonrpc_client_class_init (JsonrpcClientClass *klass)
    * in a timely manner using jsonrpc_client_reply() or
    * jsonrpc_client_reply_async().
    *
+   * Additionally, since 3.28 you may connect to the "detail" of this signal
+   * to handle a specific method call. Use the method name as the detail of
+   * the signal.
+   *
    * Since: 3.26
    */
   signals [HANDLE_CALL] =
     g_signal_new ("handle-call",
                   G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST,
+                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                   G_STRUCT_OFFSET (JsonrpcClientClass, handle_call),
                   g_signal_accumulator_true_handled, NULL, NULL,
                   G_TYPE_BOOLEAN,
@@ -458,7 +490,7 @@ jsonrpc_client_class_init (JsonrpcClientClass *klass)
   signals [NOTIFICATION] =
     g_signal_new ("notification",
                   G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST,
+                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                   G_STRUCT_OFFSET (JsonrpcClientClass, notification),
                   NULL, NULL, NULL,
                   G_TYPE_NONE,
@@ -633,8 +665,10 @@ jsonrpc_client_call_read_cb (GObject      *object,
 
       if (g_variant_dict_lookup (&dict, "method", "&s", &method_name))
         {
+          GQuark detail = g_quark_try_string (method_name);
+
           params = g_variant_dict_lookup_value (&dict, "params", NULL);
-          g_signal_emit (self, signals [NOTIFICATION], 0, method_name, params);
+          g_signal_emit (self, signals [NOTIFICATION], detail, method_name, params);
         }
 
       goto begin_next_read;
@@ -673,6 +707,7 @@ jsonrpc_client_call_read_cb (GObject      *object,
       g_autoptr(GVariant) params = NULL;
       const gchar *method_name = NULL;
       gboolean ret = FALSE;
+      GQuark detail;
 
       if (!g_variant_dict_lookup (&dict, "method", "&s", &method_name) ||
           NULL == (id = g_variant_dict_lookup_value (&dict, "id", NULL)))
@@ -689,24 +724,13 @@ jsonrpc_client_call_read_cb (GObject      *object,
       g_assert (method_name != NULL);
       g_assert (id != NULL);
 
-      g_signal_emit (self, signals [HANDLE_CALL], 0, method_name, id, params, &ret);
+      detail = g_quark_try_string (method_name);
+      g_signal_emit (self, signals [HANDLE_CALL], detail, method_name, id, params, &ret);
 
       if (ret == FALSE)
-        {
-          GVariantDict reply;
-          GVariantDict error_dict;
-
-          g_variant_dict_init (&error_dict, NULL);
-          g_variant_dict_insert (&error_dict, "code", "i", -32601);
-          g_variant_dict_insert (&error_dict, "message", "s", "The method does not exist or is not available");
-
-          g_variant_dict_init (&reply, NULL);
-          g_variant_dict_insert (&reply, "jsonrpc", "s", "2.0");
-          g_variant_dict_insert_value (&reply, "id", id);
-          g_variant_dict_insert_value (&reply, "error", g_variant_dict_end (&error_dict));
-
-          jsonrpc_output_stream_write_message_async (priv->output_stream, g_variant_dict_end (&reply), NULL, NULL, NULL);
-        }
+        jsonrpc_client_reply_error_async (self, id, JSONRPC_CLIENT_ERROR_METHOD_NOT_FOUND,
+                                          "The method does not exist or is not available",
+                                          NULL, NULL, NULL);
 
       goto begin_next_read;
     }
@@ -750,7 +774,9 @@ jsonrpc_client_call_read_cb (GObject      *object,
   g_warning ("Unhandled RPC from peer!");
 
 begin_next_read:
-  if (priv->input_stream != NULL && priv->in_shutdown == FALSE)
+  if (priv->input_stream != NULL &&
+      priv->in_shutdown == FALSE &&
+      priv->failed == FALSE)
     jsonrpc_input_stream_read_message_async (priv->input_stream,
                                              priv->read_loop_cancellable,
                                              jsonrpc_client_call_read_cb,
@@ -816,7 +842,7 @@ jsonrpc_client_call (JsonrpcClient  *self,
 
   main_context = g_main_context_ref_thread_default ();
 
-  task = g_task_new (NULL, NULL, NULL, NULL);
+  task = g_task_new (self, NULL, NULL, NULL);
   g_task_set_source_tag (task, jsonrpc_client_call);
 
   jsonrpc_client_call_async (self,
@@ -1179,6 +1205,8 @@ jsonrpc_client_close (JsonrpcClient  *self,
         g_task_return_error (task, g_error_copy (local_error));
     }
 
+  g_signal_emit (self, signals [FAILED], 0);
+
   return ret;
 }
 
@@ -1238,6 +1266,88 @@ jsonrpc_client_close_finish (JsonrpcClient  *self,
   return g_task_propagate_boolean (G_TASK (result), error);
 }
 
+static void
+jsonrpc_client_reply_error_cb (GObject      *object,
+                               GAsyncResult *result,
+                               gpointer      user_data)
+{
+  JsonrpcOutputStream *stream = (JsonrpcOutputStream *)object;
+  g_autoptr(GTask) task = user_data;
+  g_autoptr(GError) error = NULL;
+
+  g_assert (JSONRPC_IS_OUTPUT_STREAM (stream));
+  g_assert (G_IS_ASYNC_RESULT (result));
+  g_assert (G_IS_TASK (task));
+
+  if (!jsonrpc_output_stream_write_message_finish (stream, result, &error))
+    g_task_return_error (task, g_steal_pointer (&error));
+  else
+    g_task_return_boolean (task, TRUE);
+}
+
+void
+jsonrpc_client_reply_error_async (JsonrpcClient       *self,
+                                  GVariant            *id,
+                                  gint                 code,
+                                  const gchar         *message,
+                                  GCancellable        *cancellable,
+                                  GAsyncReadyCallback  callback,
+                                  gpointer             user_data)
+{
+  JsonrpcClientPrivate *priv = jsonrpc_client_get_instance_private (self);
+  g_autoptr(GTask) task = NULL;
+  g_autoptr(GError) error = NULL;
+  g_autoptr(GVariant) vreply = NULL;
+  GVariantDict reply;
+  GVariantDict error_dict;
+
+  g_return_if_fail (JSONRPC_IS_CLIENT (self));
+  g_return_if_fail (id != NULL);
+  g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));
+
+  if (message == NULL)
+    message = "An error occurred";
+
+  task = g_task_new (self, cancellable, callback, user_data);
+  g_task_set_source_tag (task, jsonrpc_client_reply_error_async);
+  g_task_set_priority (task, G_PRIORITY_LOW);
+
+  if (!jsonrpc_client_check_ready (self, &error))
+    {
+      g_task_return_error (task, g_steal_pointer (&error));
+      return;
+    }
+
+  g_variant_dict_init (&error_dict, NULL);
+  g_variant_dict_insert (&error_dict, "code", "i", code);
+  g_variant_dict_insert (&error_dict, "message", "s", message);
+
+  g_variant_dict_init (&reply, NULL);
+  g_variant_dict_insert (&reply, "jsonrpc", "s", "2.0");
+  g_variant_dict_insert_value (&reply, "id", id);
+  g_variant_dict_insert_value (&reply, "error", g_variant_dict_end (&error_dict));
+
+  vreply = g_variant_take_ref (g_variant_dict_end (&reply));
+
+  jsonrpc_output_stream_write_message_async (priv->output_stream,
+                                             vreply,
+                                             cancellable,
+                                             jsonrpc_client_reply_error_cb,
+                                             g_steal_pointer (&task));
+}
+
+gboolean
+jsonrpc_client_reply_error_finish (JsonrpcClient  *self,
+                                   GAsyncResult   *result,
+                                   GError        **error)
+{
+  g_return_val_if_fail (JSONRPC_IS_CLIENT (self), FALSE);
+  g_return_val_if_fail (G_IS_TASK (result), FALSE);
+  g_return_val_if_fail (g_task_is_valid (G_TASK (result), self), FALSE);
+
+  return g_task_propagate_boolean (G_TASK (result), error);
+}
+
 /**
  * jsonrpc_client_reply:
  * @id: (transfer none) (not nullable): the id of the message to reply
